# Author: Quentin GENOUD DUVILLARET (quentin.genoud@cern.ch)
# Scirpt to automate WRS calibration

import os
import sys
import signal
import time
import argparse
from scope import Scope
from wrs import WRS
from logger import Logger
from config import Config
from customthread import CustomThread
import threading
#import subprocess

#############
# FUNCTIONS #
#############
# Ctrl+C handler, quit immediately
def signal_handler(sig, frame):
    print("Ctrl+C captured, ending")
    quit()

def get_num_input(text, val_min, val_max):
    while True:
        value = int(input(text))
        if(value >= val_min and value <= val_max):
            break
    return value

def get_addr(text):
    while True:
        addr = input(text)
        response = os.system("ping -W 1 -c 1 " + addr + " > /dev/null")
        if(response == 0):
            break
        else:
            print("ERROR: wrong address or device is offline")
    return addr

def get_yes_no(text):
    while True:
        value = input(text + " (y/n)")
        if(value == "y" or value == "Y"):
            return True
        elif(value == "n" or value == "N"):
            return False

# Function used to measure rawDelayMM and to log the values
def get_rawDelayMM_mean_log(logger, wrs, meas_nb):
    curr_meas_nb = 0
    delayMM_sum = 0.00
    lines = ""
    logger.write_rawDelayMM_header()
    start = time.time()
    while(curr_meas_nb < meas_nb):
        curr_meas_nb += 1
        tstart = time.time()
        delayMM = wrs.get_rawDelayMM(wrs.sfp_port)
        logger.write_rawDelayMM(curr_meas_nb, delayMM)
        tdiff = (time.time() - tstart)
        delayMM_sum += delayMM
        #print("Meas {}: {}".format(curr_meas_nb, delayMM))
        twait = 1 - tdiff
        #print("Duration: " + str(tdiff) + ", Time to wait: " + str(twait))
        if(twait > 0):
            #print("Duration: " + str(tdiff) + ", Time to wait: " + str(twait))
            time.sleep(twait)
    #print("duration get raw delay MM mean log: {}".format(time.time()-start))
    return (delayMM_sum / meas_nb)

# Use threads to apply config and restart hald at the same time on GM & BC
def hald_restart(gm, bc):
    def hald_restart_gm(gm):
        gm.hald_restart()
    def hald_restart_bc(bc):
        bc.hald_restart()

    t1 = threading.Thread(target=hald_restart_gm, args=(gm,))
    t2 = threading.Thread(target=hald_restart_bc, args=(bc,))
            
    t1.start()
    t2.start()
            
    t1.join()
    t2.join()

def set_wrs_config(gm, bc):
    # Set GM as GM (if it isn't), reset alpha and delays to 0
    print("INFO: configuring GM {}: ".format(gm.address), end="", flush=True)
    gm.set_mode_GM()
    for i in range(1, 19):
        gm.set_port_master(i)
    gm.set_alpha(0)
    gm.set_port_delays(gm.sfp_port, 0, 0)
    print("OK")

    print("INFO: configuring BC {}: ".format(bc.address), end="", flush=True)
    bc.set_mode_BC()
    bc.set_alpha(0)
    bc.set_port_slave(bc.sfp_port)
    bc.set_port_delays(bc.sfp_port, 0, 0)
    print("OK")

    print("INFO: applying config on GM & BC: ", end="", flush=True)
    hald_restart(gm, bc)
    print("OK")

    print("INFO: checking WRSes configuration: ", end="", flush=True)
    gm_state = gm.is_gm();
    bc_state = bc.is_bc();
    if gm_state and bc_state:
        print("OK")
    else:
        print("FAIL: {} GM = {}, {} BC = {}".format(gm.address, gm_state, bc.address, bc_state))
        quit()

    print("INFO: waiting for GM PLL lock: ", end="", flush=True)
    gm.wait_until_locked(500)
    print("OK")


#############
# VARIABLES #
#############
## Default config ##
# Duration of each measure taken on the scope in s
meas_duration = 120
# Number of measures to be taken on the scope
meas_number = 20
# Number of measures of rawDelayMM to be averaged
delay_meas_nb = 20

## Default scope config ##
scope = Scope("MSO64B")
scope.model = "mso64b"
# horizontal config
scope.h_scale = 40e-12
scope.trigger_mode = "NORMAL"
scope.trigger_src = "CH1"
scope.trigger_slope = "RISE"
scope.trigger_level = 850e-3
# CH1 default config
scope.ch_label[0] = "PPS_GM"
scope.ch_vscale[0] = 10e-3
scope.ch_termination[0] = 50
scope.ch_bw[0] = "FULL"
scope.ch_offset[0] = 850e-3
# CH3 default config
scope.ch_label[2] = "PPS_BC"
scope.ch_vscale[2] = 10e-3
scope.ch_termination[2] = 50
scope.ch_bw[2] = "FULL"
scope.ch_offset[2] = 850e-3
# Measure1 config
scope.meas_type[0] = "DELAY"
scope.meas_src1[0] = "CH1"
scope.meas_src2[0] = "CH3"
scope.meas_riseh[0] = "1.6"
scope.meas_risem[0] = "0.85"
scope.meas_risel[0] = "0"

## WRSes used during the calibration ##
gm = WRS("calib_gm")
bc = WRS("calib_bc")

## Config parser ##
config = Config(scope, gm, bc)

## Logger ##
logger = Logger(scope, gm, bc)


#############
### START ###
#############
# Declare CTRL-C (SIGINT) signal handler
signal.signal(signal.SIGINT, signal_handler)

# Arg parser
parser = argparse.ArgumentParser(
                    prog=sys.argv[0],
                    description='Automate WRS calibration procedure',
                    epilog='')
"""
parser.add_argument('wrs_addr',
        help="Address of the White Rabbit Switch")
parser.add_argument('-p', '--port',
        required=True,
        type=int,
        choices=range(1, 19),
        help="The port which should be used to measure rawDelayMM")
parser.add_argument('-d', '--delta',
        action='store_true',
        required=False,
        help="If set, measure the short fiber delta1 value, else, user must provide it")
"""

parser.add_argument('-e', '--export',
        required=False,
        type=str,
        help="If set, export tx_phase_cal.conf from specified WRS to all others")
parser.add_argument('-t', '--t24p',
        required=False,
        type=int,
        choices=range(1, 19),
        help="Enable t24p measures and start at specified port")
parser.add_argument('-a', '--alpha',
        required=False,
        type=float,
        help="Set the value of alpha parameter on both switches then quit")
parser.add_argument('-l', '--log',
        required=False,
        type=str,
        #type=argparse.FileType('w'),
        default="measures.csv",
        help="Name of the file to store the measures. Default: measures.csv")
parser.add_argument('-n', '--number',
        required=False,
        type=int,
        help="Number of rawDelayMM measures to be taken")
parser.add_argument('-d', '--duration',
        required=False,
        type=int,
        help="Time to wait in seconds before saving the values on the scope after they have been reset")
parser.add_argument('-r', '--reboot',
        action='store_true',
        required=False,
        help="If set, reboot the BC WRS between each measure of skewPPS during fine tuning")
parser.add_argument('-i', '--ifdownup',
        action='store_true',
        required=False,
        help="If set, if down then up the BC WRS SFP port between each measure of skewPPS during fine tuning")
parser.add_argument('-c', '--config',
        required=False,
        type=str,
        #type=argparse.FileType('w'),
        help="Name of the file containing the config")

args = parser.parse_args()

# Parse config file if a name has been provided
if(args.config is not None):
    config.set_file(args.config)
    config.parse()
    scope = config.scope
    gm = config.gm
    bc = config.bc
    meas_number = config.meas_number
    meas_duration = config.meas_duration

# Override default number of measures if arg provided
if(args.number is not None):
    meas_number = args.number

# Override default duration of measures if arg provided
if(args.duration is not None):
    meas_duration = args.duration

# Enable reboot between measures if flag is set
if(args.reboot == True):
    bc.reboot_en = args.reboot
if(args.ifdownup == True):
    bc.ifdownup_en = args.ifdownup

delta = False #args.delta

# Set the name of logging file and open it
logger.set_file_name(args.log)
logger.open()

# Copy tx_phase_cal from specified switch
if(args.export is not None):
    print("INFO: scp tx_phase_cal.conf from {}: ".format(args.export))
    res = os.system("scp root@{}:/usr/wr/etc/tx_phase_cal.conf ./tx_phase_cal-{}.conf 1>/dev/null 2>/dev/null".format(args.export, args.export))
    if(res != 0):
        print("ERROR: failed to scp tx_phase_cal.conf from {}".format(args.export))
        quit()
    print("OK")
    while True:
        daddr = get_addr("Destination WRS for tx_phase_cal.conf ? ")
        res = os.system("scp ./tx_phase_cal-{}.conf root@{}:/usr/wr/etc/tx_phase_cal.conf 1>/dev/null 2>/dev/null".format(args.export, daddr))
        if(res != 0):
            print("ERROR: failed to scp tx_phase_cal.conf from {}".format(args.export))
        yn = get_yes_no("Continue export ?")
        if(yn == False):
            break
    yn = get_yes_no("Continue calibration ?")
    if(yn == False):
        quit()

## Get WRSes info ##
if(gm.check_address() == False):
    gm.address = get_addr("Enter GM address? ")
if(gm.ssh_connect() == False):
    print("ERROR: could not connect SSH to {}".format(gm.address))
    quit()
if(gm.check_sfp_port() == False):
    gm.sfp_port = get_num_input("Enter GM port number (1-18)? ", 1, 18)

if(bc.check_address() == False):
    bc.address = get_addr("Enter BC address? ")
if(bc.ssh_connect() == False):
    print("ERROR: could not connect SSH to {}".format(bc.address))
    quit()
if(bc.check_sfp_port() == False):
    bc.sfp_port = get_num_input("Enter BC port number (1-18)? ", 1, 18)

gm.print_info()
bc.print_info()

## If t24p flag, get t24p, starting from specified port (paramter provided after --t24p) ##
if(args.t24p is not None):
    t24p = [0] * 32
    print("INFO: configuring BC {} to measure t24p: ".format(bc.address), end="", flush=True)
    bc.set_mode_BC()
    for i in range(args.t24p, 19):
        bc.set_port_slave(i)
    bc.hald_restart()
    bc.is_bc()
    print("OK")
    logger.write_t24p_header()
    for i in range(args.t24p, 19):
        input("Connect port {} of {} to a valid master then press ENTER".format(i, bc.address))
        print("INFO: measuring t24p for port {}: ".format(i), end="", flush=True)
        t24p[i] = bc.get_port_t24p(i)
        logger.write_t24p(i, t24p[i])
        print("OK")
        print("INFO: writing port {} t24p ({}) in dot-config: ".format(i, t24p[i]), end="", flush=True)
        bc.set_port_t24p(i, t24p[i])
        print("OK")
        if(i < 18):
            yn = get_yes_no("Continue measuring t24p for port {} ?".format(i+1))
            if(yn == False):
                last_port = i
                break
    ynbc = get_yes_no("Apply dot-config changes on BC ?")
    yngm = get_yes_no("Apply dot-config changes on GM ?")
    if(yngm == True and ynbc == True):
        print("INFO: applying config on GM & BC: ", end="", flush=True)
        for i in range(args.t24p, last_port+1):
            gm.set_port_t24p(i, t24p[i])
        hald_restart(gm, bc)
        print("OK")
    else:
        if(ynbc == True): 
            print("INFO: applying config BC {}: ".format(bc.address), end="", flush=True)
            bc.hald_restart()
            print("OK")
        if(yngm == True):
            for i in range(args.t24p, last_port+1):
                gm.set_port_t24p(i, t24p[i])
            gm.hald_restart()
    yn = get_yes_no("Continue calibration ?")
    if(yn == False):
        quit()

# If alpha flag, set input alpha value to both GM and BC
if(args.alpha is not None):
    print("INFO: write alpha = {} in dot-config on {} and {}".format(args.alpha, gm.address, bc.address))
    gm.set_alpha(args.alpha)
    bc.set_alpha(args.alpha)
    yn = get_yes_no("Apply parameters and restart hald ?")
    if(yn == True):
        print("INFO: applying config on both WRSes: ", end="", flush=True)
        hald_restart(gm, bc)
        print("OK")
    yn = get_yes_no("Continue calibration ?")
    if(yn == False):
        quit()

# Check if a scope url/address has been provided in config file
if(scope.check_address() == False):
    yn = get_yes_no("Scope MSO64B available ?")
    if(yn == True):
        scope.address = get_addr("Enter scope address? ")

# Try connect to scope
if(scope.check_address() == True):
    scope.link_api()
    if(scope.connected == True):
        print("INFO: scope {} connected".format(scope.address))
        # Set config
        print("INFO: set scope config")
        scope.set_config()
        scope.print_info()
    else:
        print("ERROR: could not connect to scope {}".format(scope.address))
        yn = get_yes_no("Continue ?")
        if(yn == False):
            quit()

# Write header (info) in header
logger.write_header()


###################################
# 1 - Calculate delta1 and delta2 #
###################################
if(delta == False):
    delta = get_yes_no("Calculate delta1/delta2 ?")

if(delta == True):
    # Set GM as GM (if it isn't), reset alpha and delays to 0
    print("INFO: configuring GM {}: ".format(gm.address), end="", flush=True)
    gm.set_mode_GM()
    for i in range(1, 19):
        gm.set_port_master(i)
    gm.set_alpha(0)
    gm.set_port_delays(gm.sfp_port, 0, 0)
    #gm.reboot()
    #gm.hald_restart()
    print("OK")

    print("INFO: configuring BC {}: ".format(bc.address), end="", flush=True)
    bc.set_mode_BC()
    bc.set_alpha(0)
    bc.set_port_slave(bc.sfp_port)
    bc.set_port_delays(bc.sfp_port, 0, 0)
    #bc.reboot()
    #bc.hald_restart()
    print("OK")

    print("INFO: applying config on GM & BC: ", end="", flush=True)
    hald_restart(gm, bc)
    print("OK")

    """
    print("INFO: waiting for WRSes to boot: ", end="", flush=True)
    while True:
        gm_state = gm.ssh_connect()
        bc_state = bc.ssh_connect()
        if gm_state and bc_state:
            break
        time.sleep(1)
    print("OK")
    """

    print("INFO: checking WRSes configuration: ", end="", flush=True)
    gm_state = gm.is_gm();
    bc_state = bc.is_bc();
    if gm_state and bc_state:
        print("OK")
    else:
        print("FAIL: {} GM = {}, {} BC = {}".format(gm.address, gm_state, bc.address, bc_state))
        quit()

    print("INFO: waiting for GM PLL lock: ", end="", flush=True)
    gm.wait_until_locked(500)
    print("OK")

    print("Plug the short fiber between {} port {} and {} port {} then press ENTER to continue...".format(gm.address, gm.sfp_port, bc.address, bc.sfp_port), end="", flush=True)
    key = input()
    print("INFO: waiting for BC PLL lock: ", end="", flush=True)
    bc.wait_until_locked(500)
    print("OK")
    print("INFO: starting rawDelayMM measure")
    rawDelayMM_short = bc.get_rawDelayMM_mean(bc.sfp_port, delay_meas_nb)
    print("INFO: delayMM_short = {}".format(str(rawDelayMM_short)))

    print("Plug the long fiber between {} port {} and {} port {} then press ENTER to continue...".format(gm.address, gm.sfp_port, bc.address, bc.sfp_port), end="", flush=True)
    key = input()
    print("INFO: waiting for BC PLL lock: ", end="", flush=True)
    bc.wait_until_locked(500)
    print("OK")
    print("INFO: starting rawDelayMM measure")
    rawDelayMM_long = bc.get_rawDelayMM_mean(bc.sfp_port, delay_meas_nb)
    print("INFO: delayMM_long = {}".format(str(rawDelayMM_long)))

    print("Plug short + long fibers between {} port {} and {} port {} then press ENTER to continue...".format(gm.address, gm.sfp_port, bc.address, bc.sfp_port), end="", flush=True)
    key = input()
    print("INFO: waiting for BC PLL lock: ", end="", flush=True)
    bc.wait_until_locked(500)
    print("OK")
    print("INFO: starting rawDelayMM measure")
    rawDelayMM_combined = bc.get_rawDelayMM_mean(bc.sfp_port, delay_meas_nb)
    print("INFO: delayMM_combined = {}".format(str(rawDelayMM_combined)))

    delta1 = rawDelayMM_combined - rawDelayMM_long
    delta2 = rawDelayMM_combined - rawDelayMM_short
    delayTxRx = round(((rawDelayMM_short-delta1)/4)*1e12)

    print("INFO:")
    print("  - rawDelayMM (short)...: {}".format(rawDelayMM_short))
    print("  - rawDelayMM (long)... : {}".format(rawDelayMM_long))
    print("  - rawDelayMM (combined): {}".format(rawDelayMM_combined))
    print("  - delta1...............: {}".format(delta1))
    print("  - delta2...............: {}".format(delta2))
    print("  - delayTx/Rx new cal...: {}".format(delayTxRx))

else:
    yn = get_yes_no("Values of rawDelayMM for short, long and combined fibers available ?")
    if(yn == True):
        rawDelayMM_short = float(input("rawDelayMM short fiber ?"))
        rawDelayMM_long = float(input("rawDelayMM long fiber ?"))
        rawDelayMM_combined = float(input("rawDelayMM combined fiber ?"))
        delta1 = rawDelayMM_combined - rawDelayMM_long
        delta2 = rawDelayMM_combined - rawDelayMM_short
        delayTxRx = round(((rawDelayMM_short-delta1)/4)*1e12)
        delta = True
    else:
        # TODO: here at least we should measure rawDelayMM for short fiber again to be
        # able to calculate delays for a new calibrator, if needed after
        # Maybe just ask because it is not needed in case of an already calibrated calibrator xD
        delta1 = float(input("delta1 (short fiber) in seconds ?"))
        yn = get_yes_no("Get rawDelayMM short fiber to turn GM into calibrator ?")
        if(yn == True):
            set_wrs_config(gm, bc)
            print("INFO: waiting for BC PLL lock: ", end="", flush=True)
            bc.wait_until_locked(500)
            print("OK")
            print("INFO: starting rawDelayMM measure")
            rawDelayMM_short = bc.get_rawDelayMM_mean(bc.sfp_port, delay_meas_nb)
            print("INFO: delayMM_short = {}".format(str(rawDelayMM_short)))
            delayTxRx = round(((rawDelayMM_short-delta1)/4)*1e12)
            print("INFO: calculated delay Tx/Rx: {}".format(delayTxRx))
        else:
            delayTxRx = 0


#########################################
# 2 - Calculate fiber alpha coefficient #
#########################################
alpha = 0
# Values from deltas calculation are needed to calculate alpha
if(delta == True):
    yn = get_yes_no("Calculate alpha ?")
    if (yn == True):
        if(scope.connected == False):
            # Measure skewPPS1 (tPPS_slave – tPPS_master, short fiber) on scope
            skewPPS1 = float(input("skewPPS1 (tPPS_slave – tPPS_master, SHORT fiber) ?"))
            # Measure skewPPS2 (tPPS_slave – tPPS_master, long fiber) on scope
            skewPPS2 = float(input("skewPPS2 (tPPS_slave – tPPS_master, LONG fiber) ?"))
        else:
            input("Plug SHORT fiber between GM and BC, then press ENTER")
            print("INFO: waiting for BC PLL lock: ", end="", flush=True)
            bc.wait_until_locked(500)
            print("OK")
            scope.auto_adjust()
            input("Wait until scope measure stable then press ENTER")
            skewPPS1 = float(scope.api.get_mean(1))
            input("Plug LONG fiber between GM and BC, then press ENTER")
            print("INFO: waiting for BC PLL lock: ", end="", flush=True)
            bc.wait_until_locked(500)
            print("OK")
            scope.auto_adjust()
            input("Wait until scope measure stable then press ENTER")
            skewPPS2 = float(scope.api.get_mean(1))
        # Calculate alpha
        alpha = (2*(skewPPS2-skewPPS1)) / ((delta2/2) - (skewPPS2 - skewPPS1))
        alpha = round(alpha, 8)
        print("INFO:")
        print("  - skewPPS1 (short fiber): {}".format(skewPPS1))
        print("  - skewPPS2 (long fiber).: {}".format(skewPPS2))
        print("  - calulated alpha.......: {}".format(alpha))

yn = get_yes_no("Apply alpha to both WRSes ?")
if(yn == True):
    if(alpha == 0):
        alpha = float(input("Alpha value ?"))

    print("INFO: applying alpha ({}) on both WRSes: ".format(alpha), end="", flush=True)
    gm.set_alpha(alpha)
    bc.set_alpha(alpha)
    hald_restart(gm, bc)
    print("OK")


#############################################################
# 3 - Prepare GM (if new calibrator) and BC for calibration #
#############################################################
yn = get_yes_no("Apply delay Tx / Rx ({}) to turn GM {} into new calibrator ?".format(delayTxRx, gm.address))
if(yn == True):
    print("INFO: set {} port 1 Tx/Rx delays ({}): ".format(gm.address, delayTxRx), end="", flush=True)
    gm.set_port_delays(1, delayTxRx, delayTxRx)
    gm.ppsi_restart()
    print("OK")
else:
    print("INFO: leaving GM (calibrator) as it is")
    """
    input("INFO: plug the new calibrator (SFP in port 1) and press ENTER")
    gm.golden = get_yes_no("Is the calibrator the golden calibrator (old version not compatible) ?")
    if(gm.golden == False):
        gm.address = get_addr("Enter new GM (calibrator) address ?")
        gm.ssh_connect()
    """

print("INFO: set BC {} ports as slave and reset delays".format(bc.address), end="", flush=True)
for i in range(1, 19):
    bc.set_port_slave(i)
    bc.set_port_delays(i, 0, 0)
    print(".", end="", flush=True)
print(" OK")

# Apply config (alpha and possibly GM Tx/Rx delays)
print("INFO: applying BC config: ", end="", flush=True)
bc.ppsi_restart()
print("OK")

print("INFO: checking WRSes configuration: ", end="", flush=True)
if(gm.golden == True):
    gm_state = True
else:
    gm_state = gm.is_gm()
bc_state = bc.is_bc()
if gm_state and bc_state:
    print("OK")
else:
    print("FAIL: {} GM = {}, {} BC = {}".format(gm.address, gm_state, bc.address, bc_state))
    quit()

if(gm.golden == False):
    print("INFO: waiting for GM PLL lock: ", end="", flush=True)
    gm.wait_until_locked(500)
    print("OK")


#############################
# 4 - Calibrating all ports #
#############################
# Ingress / egress values for port 1, normal SFP config (Purple to Blue)
deltaTxP1_PtB = 0
deltaRxP1_PtB = 0
sfp_calib = False
while True:
    # Connect the switch to be calibrated to port 1 of the calibrator
    key = input("Plug the short fiber between port 1 of calibrator and port {} of {} then press ENTER to continue...".format(bc.sfp_port, bc.address))

    # Wait until BC is synchronized
    print("INFO: waiting for BC PLL lock: ", end="", flush=True)
    bc.wait_until_locked(500)
    print("OK")
    
    # Get Master PHY delays Tx / Rx
    print("INFO: get master PHY delay Tx / Rx from BC: ", end="", flush=True)
    (delayTx, delayRx) = bc.get_masterDelays(bc.sfp_port)
    print("OK")

    # Get rawDelayMM
    print("INFO: get rawDelayMM from BC: ", end="", flush=True)
    rawDelayMM = get_rawDelayMM_mean_log(logger, bc, delay_meas_nb)
    print("OK")

    print("INFO: port {}".format(bc.sfp_port))
    print("  - master delay Tx: {}".format(delayTx))
    print("  - master delay Rx: {}".format(delayRx))
    print("  - rawDelayMM.....: {}".format(rawDelayMM))
    
    # Get bitslide if port is > 12
    bitslide = 0
    if(bc.sfp_port > 12):
        bitslide = bc.get_bitslide(bc.sfp_port)
        print("  - bitslide = {}".format(bitslide))

    # Calculate average delta TxS / RxS
    deltaTxRx = ((rawDelayMM - delayTx - delayRx - bitslide - delta1) / 2) * 1e12
    deltaTx = int(round(deltaTxRx))
    deltaRx = int(round(deltaTxRx))

    # Apply delta to the port and apply config
    print("INFO: applying delayTx = {} and delayRx = {} on port {}: ".format(deltaTx, deltaRx, bc.sfp_port), end="", flush=True)
    bc.set_port_delays(bc.sfp_port, deltaTx, deltaRx)
    bc.ppsi_restart()
    bc.wait_until_locked(500)
    time.sleep(1)
    print("OK")

    # First iteration (coarse), measure skewPPS and update delays
    if(scope.connected == True):
        scope.auto_adjust()
        time.sleep(5)
        skew = float(scope.api.get_mean(1))
        nsamples = int(scope.api.get_nsamples(1))
        print("INFO: first iteration (coarse): measured mean skewPPS = {} s ({} samples)".format(skew, nsamples))
    else:
        skew = float(input("skewPPS ?"))

    deltaTx = int(round((deltaTx - (skew * 1e12))))
    deltaRx = int(round((deltaRx + (skew * 1e12))))

    print("INFO: writing port {} delays in dot config:".format(bc.sfp_port))
    print("  - deltaTx (egress latency) : {}".format(deltaTx))
    print("  - deltaRx (ingress latency): {}".format(deltaRx))
    bc.set_port_delays(bc.sfp_port, deltaTx, deltaRx)

    # if reboot mode is not enable, apply parameters here
    if(bc.reboot_en == False):
        print("INFO: applying config on BC: ", end="", flush=True)
        bc.ppsi_restart()
        bc.wait_until_locked(500)
        time.sleep(1)
        print("OK")

    logger.write_meas_header()

    # Second iteration (fine), measure skewPPS, reboot BC, log measures and update delays
    print("INFO: starting second iteration (fine tune), {} measures of {} s".format(meas_number, meas_duration))
    curr_meas_nb = 0
    total_skew = 0
    while(curr_meas_nb < meas_number):
        curr_meas_nb += 1
        if(bc.reboot_en == True):
            print("INFO: rebooting BC {}: ".format(bc.address), end="", flush=True)
            while(bc.is_up() == True):
                bc.reboot()
            print("down", end="", flush=True)
            while(bc.is_up() == False):
                pass
            print(", up", end="", flush=True)
            while(bc.ssh_connect() == False):
                time.sleep(1)
            print(", SSH connected")
        elif(bc.ifdownup_en == True):
            print("INFO: ifdownup BC port {}: ".format(bc.sfp_port), end="", flush=True)
            bc.ifdownup()
            print("OK")

        print("INFO: waiting for BC PLL lock: ", end="", flush=True)
        bc.wait_until_locked(500)
        print("OK")

        if(scope.connected == True):
            scope.auto_adjust()
            print("INFO: waiting {} s".format(meas_duration))
            time.sleep(meas_duration)
            skew = float(scope.api.get_mean(1))
            total_skew += skew
            nsamples = int(scope.api.get_nsamples(1))
            logger.write_measure(curr_meas_nb)
            print("INFO: -- saving measure #{} (skewPPS mean = {}, {} samples, average: {}) --".format(curr_meas_nb, skew, nsamples, total_skew/curr_meas_nb))
            #print("INFO: first iteration (coarse) skewPPS mean from scope: {} s ({} samples)".format(skew, nsamples))
        else:
            skew = float(input("skewPPS ?"))

    skew = total_skew / meas_number
    print("INFO: {} measures taken after reboot, mean skew: {} s".format(meas_number, skew))

    deltaTx = int(round((deltaTx - (skew * 1e12))))
    deltaRx = int(round((deltaRx + (skew * 1e12))))

    print("INFO: writing dot-config:")
    print("  - deltaTx (egress latency) : {}".format(deltaTx))
    print("  - deltaRx (ingress latency): {}".format(deltaRx))
    bc.set_port_delays(bc.sfp_port, deltaTx, deltaRx)
    logger.write_egress_ingress(deltaTx, deltaRx)

    yn = get_yes_no("Apply latest deltaTx ({}) / deltaRx ({}) on port {} ?".format(deltaTx, deltaRx, bc.sfp_port))
    if(yn == True):
        print("INFO: applying delays on port {}: ".format(bc.sfp_port), end="", flush=True)
        bc.ppsi_restart()
        bc.wait_until_locked(500)
        print("OK")

    # If current port is port 18, then calibration is over
    if(bc.sfp_port == 18):
        break
    else:
        # Only on port 1, calculate SFPs Tx/Rx correction values
        if(bc.sfp_port == 1):
            # Port 1 calibrated, we assume that:
            # - it was with default SFP type (BlueOptics)
            # - the SFPs were in the "normal" config: purple to blue
            # Save calibration values
            if(deltaTxP1_PtB == 0 and deltaRxP1_PtB == 0):
                deltaTxP1_PtB = deltaTx
                deltaRxP1_PtB = deltaRx
            # We already have the calibration values for default SFPs normal config
            # So just calculate the values for the current SFP config
            else:
                print("INFO: alternative SFP config, calibration values:")
                print("  - SFP Tx correction: {}".format(deltaTx-deltaTxP1_PtB))
                print("  - SFP Rx correction: {}".format(deltaRx-deltaRxP1_PtB))
                logger.write_sfp_correction_header()
                logger.write_sfp_correction(deltaTx-deltaTxP1_PtB, deltaRx-deltaRxP1_PtB)
                ### TODO: SAVE SFP CORRECTION VALUES IN DOT-CONFIG 
            
            yn = get_yes_no("Calibrate other SFP config (SFP correction value) ?")
            if(yn == True):
                bc.set_port_delays(bc.sfp_port, 0, 0)
                bc.ppsi_restart()
                bc.wait_until_locked(500)
                input("Plug the SFPs to be calibrated on port 1, then press ENTER")
            else:
                bc.sfp_port = bc.sfp_port + 1
                yn = get_yes_no("Continue calibration for port {} ?".format(bc.sfp_port))
                if(yn == False):
                    break
        else:
            bc.sfp_port = bc.sfp_port + 1
            yn = get_yes_no("Continue calibration for port {} ?".format(bc.sfp_port))
            if(yn == False):
                break

quit()
