import os
import sys
from wrs import WRS
from scope import Scope

# Custom string class redefining "==" operator to be used in "match" 
class CfgStr(str):
    def __eq__(self, other):
        return self.__contains__(other)

class Config:
    # Section names in config file
    cfg_file: str
    scope: Scope
    gm: WRS
    bc: WRS
    meas_number = 20
    meas_duration = 60

    # Init: open the config file provided as argument 
    def __init__(self, scope, wrs_gm, wrs_bc):
        self.scope = scope
        self.gm = wrs_gm
        self.bc = wrs_bc

    def set_file(self, cfg_file):
        self.cfg_file = cfg_file

    # Get all lines contained in a section
    def get_section_lines(self, lines, first_line_index) -> [str]:
        section_lines = []
        for line in lines[first_line_index:]:
            # If there is ':' in a line, that means that we are reading the title of another section
            if(line.find(':') > 0):
                break
            elif(line[0] != '\n'):
                line.replace(" ", "")
                section_lines += [line.strip()]
        return section_lines

    # Parse a section describing a scope
    def parse_scope_section(self, scope, section_lines, first_line_nb) -> Scope:
        line_nb=first_line_nb
        for sline in section_lines:
            line_nb += 1
            #print("parse_scope_section: parsing line {}: {}".format(line_nb, sline))
            # If '=' char is in the line, then split line into property / value
            if(sline.find('=') > 0):
                prop = sline.split('=')[0]
                value = sline.split('=')[1]
                if(prop.find("ch") > -1 and prop.find("name") > -1 and len(value) > 0):
                    scope.nb_ch +=1
                #print("Parsing line {}, prop={}, value={}".format(line_nb, prop, value))
                propstr = CfgStr(prop)
                match propstr:
                    case "name":
                        scope.name = value
                    case "model":
                        scope.model = value
                    case "address":
                        scope.address = value
                    case "ch":
                        for i in range(4):
                            if(prop.find("ch"+str(i+1)) > -1):
                                #print("ch" + str(i+1) + " detected in config: " + prop + " = " + value)
                                match propstr:
                                    case "label":
                                        scope.ch_label[i] = value
                                    case "vscale":
                                        scope.ch_vscale[i] = value
                                    case "termination":
                                        scope.ch_termination[i] = value
                                    case "bw":
                                        scope.ch_bw[i] = value
                                    case "offset":
                                        scope.ch_offset[i] = value
                                    case _:
                                        print("WARNING: CH{}, prop: {} not recognized".format(i+1, prop))
                    case "h_scale":
                        scope.h_scale = value
                    case "trigger_mode":
                        scope.trigger_mode = value
                    case "trigger_src":
                        scope.trigger_src = value
                    case "trigger_level":
                        scope.trigger_level = value
                    case "trigger_slope":
                        scope.trigger_slope = value
                    case "meas":
                        for i in range(3):
                            if(prop.find("meas"+str(i+1)) > -1):
                                #print("meas" + str(i+1) + " detected in config: " + prop + " = " + value)
                                match propstr:
                                    case "type":
                                        scope.meas_type[i] = value
                                    case "src1":
                                        scope.meas_src1[i] = value
                                    case "src2":
                                        scope.meas_src2[i] = value
                                    case "riseh":
                                        scope.meas_riseh[i] = value
                                    case "risem":
                                        scope.meas_risem[i] = value
                                    case "risel":
                                        scope.meas_risel[i] = value
                                    case _:
                                        print("meas" + str(i+1) + " prop: " + prop + " not recognized ")
                    case _:
                        print("WARNING: line {}: property \"{}\" not recognized".format(line_nb, prop))
            # No '=' char in line, this can't be parsed
            elif(sline[0] != "#"):
                print("WARNING: line {}: \"{}\" does not assign a property".format(line_nb, sline))

        return scope

    # Parse a section describing a WRS
    def parse_switch_section(self, switch, section_lines, first_line_nb) -> WRS:
        line_nb=first_line_nb
        for sline in section_lines:
            line_nb += 1
            # If '=' char is in the line, then split line into property / value
            if(sline.find('=') > 0):
                prop = sline.split('=')[0]
                value = sline.split('=')[1]
                match prop:
                    case "name":
                        switch.name = value
                    case "address":
                        switch.address = value
                    case "password":
                        switch.password = value
                    case "sfp_port":
                        switch.sfp_port = int(value)
                    case "golden":
                        if(value == "true"):
                            switch.golden = True
                        else:
                            switch.golden = False
                    case "reboot":
                        if(value == "true"):
                            switch.reboot_en = True
                        else:
                            switch.reboot_en = False
                    case "ifdownup":
                        if(value == "true"):
                            switch.ifdownup_en = True
                        else:
                            switch.ifdownup_en = False
                    case _:
                        print("WARNING: line {}: property \"{}\" not recognized".format(line_nb, prop))
            # No '=' char in line, this can't be parsed
            elif(sline[0] != "#"):
                print("WARNING: line {}: \"{}\" does not assign a property".format(line_nb, sline))

        return switch
    
    # Parse file
    def parse(self):
        section_names = ["scope1", "scope2", "gm", "bc", "config"]
        # Try to open config file in read only
        print("INFO: trying to open config file \"{}\": ".format(self.cfg_file), end="", flush=True)
        try:
            f = open(self.cfg_file, "r")
            print("OK")
        except:
            print("ERROR (file does not exist)")
            quit()
        
        lines = f.readlines()
        f.close()

        line_index = 0
        section_lines = []
        section_found = False

        # For each line in the config file
        for line in lines:
            line_index += 1

            # Parse each line which is not starting by '#' or '\n'
            if(line[0] != '\n' and line[0] != '#'):
                # Remove white spaces and '\n'
                line.replace(" ", "")
                line = line.strip()

                # If there is ':' in the line, this is a section
                if(line.find(':') > 0):
                    section_found = False
                    # Search if the section is defined in section_names
                    for sname in section_names:
                        if(line.find(sname) >= 0):
                            print("INFO: Found section: " + sname)
                            section_found = True
                            match sname:
                                case "scope1":
                                    section_lines = self.get_section_lines(lines, line_index)
                                    self.scope = self.parse_scope_section(self.scope, section_lines, line_index)
                                
                                case "gm":
                                    section_lines = self.get_section_lines(lines, line_index)
                                    self.gm = self.parse_switch_section(self.gm, section_lines, line_index)

                                case "bc":
                                    section_lines = self.get_section_lines(lines, line_index)
                                    self.bc = self.parse_switch_section(self.bc, section_lines, line_index)

                                case "config":
                                    section_lines = self.get_section_lines(lines, line_index)
                                    line_nb=line_index
                                    for sline in section_lines:
                                        line_nb += 1
                                        # If '=' char is in the line, then split line into property / value
                                        if(sline.find('=') > 0):
                                            prop = sline.split('=')[0]
                                            value = sline.split('=')[1]
                                            match prop:
                                                case "meas_number":
                                                    self.meas_number = int(value)
                                                case "meas_duration":
                                                    self.meas_duration = int(value)
                                                case _:
                                                    print("WARNING: line {}: property \"{}\" not recognized".format(line_nb, prop))
                                        # No '=' char in line, this can't be parsed
                                        elif(sline[0] != "#"):
                                            print("WARNING: line {}: \"{}\" does not assign a property".format(line_nb, sline))
                                        
                            break
                    if(section_found == False):
                        print("WARNING: line {}: section \"{}\" does not exist".format(line_index, line.split(':')[0]))
